"""Support code for generating code from abstract syntax trees."""

from __future__ import annotations

import ast
from copy import deepcopy
from typing import TYPE_CHECKING
from typing import ClassVar


if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Union
    _NodeTransform = Callable[[ast.AST], Union[ast.AST, None]]


__docformat__ = 'restructuredtext en'


def parse(source, mode='eval'):
    return compile(source, '', mode, ast.PyCF_ONLY_AST)


def load(name):
    return ast.Name(id=name, ctx=ast.Load())


def store(name):
    return ast.Name(id=name, ctx=ast.Store())


def param(name):
    return ast.Name(id=name, ctx=ast.Param())


def subscript(name, value, ctx):
    return ast.Subscript(
        value=value,
        slice=ast.Index(value=ast.Str(s=name)),
        ctx=ctx,
    )


class Node(ast.AST):
    """AST baseclass that gives us a convenient initialization
    method. We explicitly declare and use the ``_fields`` attribute."""

    _fields: ClassVar[tuple[str, ...]] = ()

    def __init__(self, *args, **kwargs) -> None:
        assert isinstance(self._fields, tuple)
        self.__dict__.update(kwargs)
        for name, value in zip(self._fields, args):
            setattr(self, name, value)

    def __repr__(self) -> str:
        """Poor man's single-line pretty printer."""

        name = type(self).__name__
        return '<{}{} at {:x}>'.format(
            name,
            "".join(
                " {}={!r}".format(name, getattr(self, name, "\"?\""))
                for name in self._fields
            ),
            id(self),
        )


class Builtin(Node):
    """Represents a Python builtin.

    Used when a builtin is used internally by the compiler, to avoid
    clashing with a user assignment (e.g. ``help`` is a builtin, but
    also commonly assigned in templates).
    """

    _fields = "id", "ctx"

    ctx = ast.Load()


class Symbol(Node):
    """Represents an importable symbol."""

    _fields = "value",


class Static(Node):
    """Represents a static value."""

    _fields = "value", "name"

    name = None


class Comment(Node):
    _fields = "text",


class TokenRef(Node):
    """Represents a source-code token reference."""

    _fields = "token",


class NodeTransformerBase(ast.NodeTransformer):
    def __init__(self, transform: _NodeTransform):
        self.transform = transform

    def apply_transform(self, node: ast.AST) -> ast.AST:
        result = self.transform(node)
        if result is not None:
            return result
        return node


class NameLookupRewriteVisitor(NodeTransformerBase):
    def __init__(self, transform: _NodeTransform):
        self.scopes: list[set[str]] = [set()]
        super().__init__(transform)

    def __call__(self, node) -> ast.AST:
        clone = deepcopy(node)
        return self.visit(clone)

    def visit_arg(self, node) -> ast.AST:
        scope = self.scopes[-1]
        scope.add(node.arg)
        return node

    def visit_Name(self, node) -> ast.AST:
        scope = self.scopes[-1]
        if isinstance(node.ctx, ast.Param):
            scope.add(node.id)
            return node
        if node.id not in scope:
            node = self.apply_transform(node)
        return node

    def visit_FunctionDef(self, node) -> ast.AST:
        self.scopes[-1].add(node.name)
        return super().generic_visit(node)

    def visit_alias(self, node) -> ast.AST:
        name = node.asname if node.asname is not None else node.name
        self.scopes[-1].add(name)
        return super().generic_visit(node)

    def visit_Lambda(self, node) -> ast.AST:
        self.scopes.append(set())
        try:
            return super().generic_visit(node)
        finally:
            self.scopes.pop()


class ItemLookupOnAttributeErrorVisitor(NodeTransformerBase):
    def visit_Attribute(self, node) -> ast.AST:
        node = self.apply_transform(node)
        return self.generic_visit(node)



from __future__ import annotations

import contextlib
import functools
import importlib.resources
import logging
import os
import posixpath
import py_compile
import shutil
import sys
import tempfile
import warnings
from importlib.machinery import SourceFileLoader
from threading import RLock
from typing import TYPE_CHECKING
from typing import Any
from typing import cast

from chameleon.utils import encode_string


if TYPE_CHECKING:
    from collections.abc import Callable
    from collections.abc import Iterator
    from collections.abc import Sequence
    from pathlib import Path
    from typing import Protocol
    from typing import TypeVar

    from .template import BaseTemplate

    _F = TypeVar('_F', bound=Callable[..., Any])
    _TemplateT = TypeVar('_TemplateT', bound='BaseTemplate')

    class _SupportsRmtree(Protocol):
        def rmtree(self, path: str, /) -> object: ...


lock = RLock()
acquire_lock = lock.acquire
release_lock = lock.release
del lock

log = logging.getLogger('chameleon.loader')


def cache(func: _F) -> _F:
    def load(self: Any, *args: Any, **kwargs: Any) -> Any:
        template = self.registry.get(args)
        if template is None:
            self.registry[args] = template = func(self, *args, **kwargs)
        return template
    return cast('_F', load)


@contextlib.contextmanager
def import_package_resource(name: str) -> Iterator[Path]:
    # FIXME: we should restrict ourselves to the Traversable protocol
    #        but for right now we assume that we always get a simple
    #        filesystem resource loader which returns pathlib.Path objects
    path: Path = importlib.resources.files(name)  # type: ignore[assignment]
    yield path
    if hasattr(path.root, 'close'):
        path.root.close()


class TemplateLoader:
    """Template loader class.

    To load templates using relative filenames, pass a sequence of
    paths (or a single path) as ``search_path``.

    To apply a default filename extension to inputs which do not have
    an extension already (i.e. no dot), provide this as
    ``default_extension`` (e.g. ``'.pt'``).

    Additional keyword-arguments will be passed on to the template
    constructor.
    """

    default_extension: str | None = None
    registry: dict[tuple[Any, ...], Any]

    def __init__(
        self,
        search_path: Sequence[str] | str | None = None,
        default_extension: str | None = None,
        **kwargs: Any
    ) -> None:

        if search_path is None:
            search_path = []
        if isinstance(search_path, str):
            search_path = [search_path]
        if default_extension is not None:
            self.default_extension = ".%s" % default_extension.lstrip('.')
        self.search_path = search_path
        self.registry = {}
        self.kwargs = kwargs

    @cache
    def load(
        self,
        spec: str,
        cls: type[_TemplateT] = None  # type: ignore[assignment]
    ) -> _TemplateT:

        if cls is None:
            raise ValueError("Unbound template loader.")

        spec = spec.strip()

        if self.default_extension is not None and '.' not in spec:
            spec += self.default_extension

        package_name = None

        if not os.path.isabs(spec):
            if ':' in spec:
                package_name, spec = spec.split(':', 1)
            else:
                for path in self.search_path:
                    if not os.path.isabs(path) and ':' in path:
                        package_name, path = path.split(':', 1)
                        with import_package_resource(package_name) as files:
                            if files.joinpath(path).joinpath(spec).exists():
                                spec = posixpath.join(path, spec)
                                break
                    else:
                        path = os.path.join(path, spec)
                        if os.path.exists(path):
                            package_name = None
                            spec = path
                            break
                else:
                    raise ValueError("Template not found: %s." % spec)

        return cls(
            spec,
            search_path=self.search_path,
            package_name=package_name,
            **self.kwargs
        )

    def bind(
        self,
        cls: type[_TemplateT]
    ) -> Callable[[str], _TemplateT]:
        return functools.partial(self.load, cls=cls)


class MemoryLoader:
    def build(self, source: str, filename: str) -> dict[str, Any]:
        code = compile(source, filename, 'exec')
        env: dict[str, Any] = {}
        exec(code, env)
        return env

    def get(self, name: str) -> None:
        return None


class ModuleLoader:
    def __init__(self, path: str, remove: bool = False) -> None:
        self.path = path
        self.remove = remove

    def __del__(self, shutil: _SupportsRmtree = shutil) -> None:
        if not self.remove:
            return
        try:
            shutil.rmtree(self.path)
        except BaseException:
            warnings.warn(
                "Could not clean up temporary file path: %s" % (self.path,)
            )

    def get(self, filename: str) -> dict[str, Any] | None:
        path = os.path.join(self.path, filename)
        if os.path.exists(path):
            log.debug("loading module from cache: %s." % filename)
            base, ext = os.path.splitext(filename)
            return self._load(base, path)
        else:
            log.debug('cache miss: %s' % filename)
            return None

    def build(self, source: str, filename: str) -> dict[str, Any]:
        acquire_lock()
        try:
            base, ext = os.path.splitext(filename)
            name = os.path.join(self.path, base + ".py")

            log.debug("writing source to disk (%d bytes)." % len(source))
            fd, fn = tempfile.mkstemp(
                prefix=base, suffix='.tmp', dir=self.path)
            temp = os.fdopen(fd, 'wb')
            encoded = source.encode('utf-8')
            header = encode_string("# -*- coding: utf-8 -*-" + "\n")

            try:
                try:
                    temp.write(header)
                    temp.write(encoded)
                finally:
                    temp.close()
            except BaseException:
                os.remove(fn)
                raise

            os.rename(fn, name)
            log.debug("compiling %s into byte-code..." % filename)
            py_compile.compile(name)

            return self._load(base, name)
        finally:
            release_lock()

    def _load(self, base: str, filename: str) -> dict[str, Any]:
        acquire_lock()
        try:
            module = sys.modules.get(base)
            if module is None:
                module = SourceFileLoader(base, filename).load_module()
        finally:
            release_lock()

        return module.__dict__



import os
import sys
import tempfile
import unittest
import zipimport
from pathlib import Path


class LoadTests:
    def _makeOne(self, *args, **kwargs):
        klass = self._getTargetClass()
        return klass(*args, **kwargs)

    def _getTargetClass(self):
        from chameleon.loader import TemplateLoader
        return TemplateLoader

    def test_load_relative(self):
        import os
        here = os.path.join(os.path.dirname(__file__), "inputs")
        loader = self._makeOne(search_path=[here])
        result = self._load(loader, 'hello_world.pt')
        self.assertEqual(
            result.filename,
            os.path.join(here, 'hello_world.pt'))

    def test_load_relative_default_extension(self):
        import os
        here = os.path.join(os.path.dirname(__file__), "inputs")
        loader = self._makeOne([here], ".pt")
        result = self._load(loader, 'hello_world')
        self.assertEqual(
            result.filename,
            os.path.join(here, 'hello_world.pt'))

    def test_consecutive_loads(self):
        import os
        here = os.path.join(os.path.dirname(__file__), "inputs")
        loader = self._makeOne(search_path=[here])

        self.assertTrue(
            self._load(loader, 'hello_world.pt') is
            self._load(loader, 'hello_world.pt'))

    def test_load_relative_badpath_in_searchpath(self):
        import os
        here = os.path.join(os.path.dirname(__file__), "inputs")
        loader = self._makeOne(search_path=[os.path.join(here, 'none'), here])
        result = self._load(loader, 'hello_world.pt')
        self.assertEqual(
            result.filename,
            os.path.join(here, 'hello_world.pt'))

    def test_load_abs(self):
        import os
        here = os.path.join(os.path.dirname(__file__), "inputs")
        loader = self._makeOne()
        abs = os.path.join(here, 'hello_world.pt')
        result = self._load(loader, abs)
        self.assertEqual(result.filename, abs)

    def test_load_egg(self):
        self._test_load_package("bdist_egg", ".egg")

    def test_load_wheel(self):
        self._test_load_package("bdist_wheel", ".whl")

    def _test_load_package(self, command, pkg_extension):
        with tempfile.TemporaryDirectory() as tmpdir:
            pkg_name = 'chameleon_test_pkg'
            basedir = Path(tmpdir) / pkg_name
            pkgdir = basedir / 'src' / pkg_name
            templatesdir = pkgdir / 'templates'
            templatesdir.mkdir(parents=True)

            olddir = os.getcwd()
            os.chdir(basedir)

            try:
                with open('MANIFEST.in', 'w') as f:
                    f.write('recursive-include src *.pt')

                pkgdir.joinpath('__init__.py').touch()
                with templatesdir.joinpath('test.pt').open('w') as f:
                    f.write("<html><body>${content}</body></html>")
                with templatesdir.joinpath('macro1.pt').open('w') as f:
                    f.write(
                        f'<html metal:use-macro="'
                        f'load: {pkg_name}:templates/test.pt" />'
                    )
                with templatesdir.joinpath('macro2.pt').open('w') as f:
                    f.write(
                        '<html metal:use-macro="load: test.pt" />'
                    )

                from setuptools import find_packages
                from setuptools import setup

                setup(
                    name=pkg_name,
                    version="1.0",
                    packages=find_packages('src'),
                    package_dir={'': 'src'},
                    include_package_data=True,
                    script_args=[command],
                )
            finally:
                os.chdir(olddir)

            (package_path,) = basedir.glob('dist/*' + pkg_extension)

            importer = zipimport.zipimporter(str(package_path))
            importer.load_module(pkg_name)

            try:
                self._test_pkg(pkg_name)
            finally:
                # Manually clean up archive.
                # See https://github.com/python/cpython/issues/87319.
                os.unlink(importer.archive)

                # Remove imported module.
                sys.modules.pop(pkg_name, None)

    def _test_pkg(self, pkg_name):
        loader = self._makeOne(auto_reload=True)
        # we use auto_reload to trigger a call of mtime
        result = self._load(
            loader, f'{pkg_name}:templates/test.pt')
        self.assertIsNone(result._v_last_read)
        output = result(content='foo')
        self.assertIsNotNone(result._v_last_read)
        old_v_last_read = result._v_last_read
        self.assertIn("foo", output)
        # make sure the template isn't recooked
        output = result(content='bar')
        self.assertEqual(result._v_last_read, old_v_last_read)
        macro1 = self._load(loader, f'{pkg_name}:templates/macro1.pt')
        macro1_output = macro1(content='bar')
        self.assertEqual(output, macro1_output)
        macro2 = self._load(loader, f'{pkg_name}:templates/macro2.pt')
        macro2_output = macro2(content='bar')
        self.assertEqual(output, macro2_output)


class LoadPageTests(unittest.TestCase, LoadTests):
    def _load(self, loader, spec):
        from chameleon.zpt import template
        return loader.load(spec, template.PageTemplateFile)


class ZPTLoadPageTests(unittest.TestCase, LoadTests):
    def _getTargetClass(self):
        from chameleon.zpt.loader import TemplateLoader
        return TemplateLoader

    def _load(self, loader, spec):
        return loader.load(spec)


class ModuleLoadTests(unittest.TestCase):
    def _makeOne(self, *args, **kwargs):
        from chameleon.loader import ModuleLoader
        return ModuleLoader(*args, **kwargs)

    def test_build(self):
        import tempfile
        path = tempfile.mkdtemp()
        loader = self._makeOne(path)
        source = "def function(): return %r" % "\xc3\xa6\xc3\xb8\xc3\xa5"

        module = loader.build(source, "test.xml")
        result1 = module['function']()
        d = {}
        code = compile(source, 'test.py', 'exec')
        exec(code, d)
        result2 = d['function']()
        self.assertEqual(result1, result2)

        import os
        self.assertTrue("test.py" in os.listdir(path))

        import shutil
        shutil.rmtree(path)


class ZPTLoadTests(unittest.TestCase):
    def _makeOne(self, *args, **kwargs):
        import os
        here = os.path.join(os.path.dirname(__file__), "inputs")
        from chameleon.zpt import loader
        return loader.TemplateLoader(here, **kwargs)

    def test_load_xml(self):
        loader = self._makeOne()
        template = loader.load("hello_world.pt", "xml")
        from chameleon.zpt.template import PageTemplateFile
        self.assertTrue(isinstance(template, PageTemplateFile))

    def test_load_text(self):
        loader = self._makeOne()
        template = loader.load("hello_world.txt", "text")
        from chameleon.zpt.template import PageTextTemplateFile
        self.assertTrue(isinstance(template, PageTextTemplateFile))

    def test_load_getitem_gets_xml_file(self):
        loader = self._makeOne()
        template = loader["hello_world.pt"]
        from chameleon.zpt.template import PageTemplateFile
        self.assertTrue(isinstance(template, PageTemplateFile))



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any
from typing import Protocol
from typing import TypedDict


if TYPE_CHECKING:
    from collections.abc import Callable
    from collections.abc import Collection

    from chameleon.tokenize import Token


class Tokenizer(Protocol):
    def __call__(
        self,
        body: str,
        filename: str | None = None
    ) -> Token: ...


class TranslationFunction(Protocol):
    def __call__(
        self,
        msgid: str,
        *,
        domain: str | None = None,
        mapping: dict[str, Any] | None = None,
        default: str | None = None,
        context: str | None = None,
        target_language: str | None = None
    ) -> str: ...


class PageTemplateConfig(TypedDict, total=False):
    auto_reload: bool
    default_expression: str
    encoding: str
    boolean_attributes: Collection[str]
    translate: TranslationFunction
    implicit_i18n_translate: bool
    implicit_i18n_attributes: set[str]
    on_error_handler: Callable[[BaseException], object]
    strict: bool
    trim_attribute_space: bool
    restricted_namespace: bool
    tokenizer: Tokenizer
    value_repr: Callable[[object], str]
    default_marker: Any



from __future__ import annotations

from chameleon.namespaces import XML_NS
from chameleon.namespaces import XMLNS_NS
from chameleon.parser import ElementParser
from chameleon.tokenize import iter_text
from chameleon.tokenize import iter_xml


class ElementProgram:
    DEFAULT_NAMESPACES = {
        'xmlns': XMLNS_NS,
        'xml': XML_NS,
    }

    tokenizers = {
        'xml': iter_xml,
        'text': iter_text,
    }

    restricted_namespace = True

    def __init__(
        self, source, mode="xml", filename=None, tokenizer=None
    ) -> None:
        if tokenizer is None:
            tokenizer = self.tokenizers[mode]
        tokens = tokenizer(source, filename)
        parser = ElementParser(
            tokens, self.DEFAULT_NAMESPACES, self.restricted_namespace
        )

        self.body = []

        for kind, args in parser:
            node = self.visit(kind, args)
            if node is not None:
                self.body.append(node)

    def visit(self, kind, args):
        visitor = getattr(self, "visit_%s" % kind)
        return visitor(*args)



from __future__ import annotations

import builtins
import re
import textwrap
import types
from ast import AST
from ast import Assign
from ast import Constant
from ast import Expr
from ast import FunctionDef
from ast import Import
from ast import ImportFrom
from ast import Module
from ast import NodeTransformer
from ast import alias
from ast import unparse

from chameleon.astutil import Builtin
from chameleon.astutil import Symbol
from chameleon.astutil import load
from chameleon.astutil import parse
from chameleon.astutil import store
from chameleon.exc import CompilationError


reverse_builtin_map = {}
for name, value in builtins.__dict__.items():
    try:
        hash(value)
    except TypeError:
        continue

    reverse_builtin_map[value] = name


def template(
    source,
    mode='exec',
    is_func: bool = False,
    func_args=(),
    func_defaults=(),
    **kw,
):
    def wrapper(*vargs, **kwargs):
        symbols = dict(zip(args, vargs + defaults))
        symbols.update(kwargs)

        class Transformer(NodeTransformer):
            def visit_FunctionDef(self, node) -> AST:
                name = symbols.get(node.name, self)
                if name is self:
                    return self.generic_visit(node)

                return FunctionDef(
                    name=name,
                    args=node.args,
                    body=list(map(self.visit, node.body)),
                    decorator_list=getattr(node, "decorator_list", []),
                    lineno=None,
                )

            def visit_Name(self, node) -> AST:
                value = symbols.get(node.id, self)
                if value is self:
                    if node.id == 'None' or \
                       getattr(builtins, node.id, None) is not None:
                        return Builtin(node.id)
                    return node

                if isinstance(value, type) or value in reverse_builtin_map:
                    name = reverse_builtin_map.get(value)
                    if name is not None:
                        return Builtin(name)
                    return Symbol(value)

                if isinstance(value, str):
                    value = load(value)

                return value

        expr = parse(textwrap.dedent(source), mode=mode)

        Transformer().visit(expr)
        return expr.body

    assert isinstance(source, str)
    defaults = func_defaults
    args = func_args
    if is_func:
        return wrapper
    else:
        return wrapper(**kw)


class TemplateCodeGenerator(NodeTransformer):
    """Generate code from AST tree.

    The syntax tree has been extended with internal nodes. We first
    transform the tree to process the internal nodes before generating
    the code string.
    """

    names = ()

    def __init__(self, tree):
        self.comments = []
        self.defines = {}
        self.imports = {}

        # Run transform.
        tree = self.visit(tree)

        # Generate code.
        code = unparse(tree)

        # Fix-up comments.
        comments = iter(self.comments)
        code = re.sub(
            r'^(\s*)\.\.\.$',
            lambda m: "\n".join(
                (m.group(1) + "#" + line)
                for line in next(comments).replace("\r", "\n").split("\n")
            ),
            code,
            flags=re.MULTILINE
        )

        self.code = code

    def define(self, name, node):
        assert node is not None
        value = self.defines.get(name)

        if value is node:
            pass
        elif value is None:
            self.defines[name] = node
        else:
            raise CompilationError(
                "Duplicate symbol name for define.", name)

        return load(name)

    def require(self, value):
        node = self.imports.get(value)
        if node is None:
            # we come up with a unique symbol based on the class name
            name = (
                "_%s"
                % getattr(value, '__name__', str(value)).rsplit('.', 1)[-1]
            )
            node = load(name)
            self.imports[value] = store(node.id)

        return node

    def visit_Module(self, module) -> AST:
        assert isinstance(module, Module)
        module = super().generic_visit(module)
        preamble: list[AST] = []

        for name, node in self.defines.items():
            assignment = Assign(targets=[store(name)], value=node, lineno=None)
            preamble.append(assignment)

        for value, node in self.imports.items():
            stmt: AST

            if isinstance(value, types.ModuleType):
                stmt = Import(
                    names=[alias(name=value.__name__, asname=node.id)])
            elif hasattr(value, '__name__'):
                path = reverse_builtin_map.get(value)
                if path is None:
                    path = value.__module__
                    name = value.__name__
                stmt = ImportFrom(
                    module=path,
                    names=[alias(name=name, asname=node.id)],
                    level=0,
                )
            else:
                raise TypeError(value)

            preamble.append(stmt)

        preamble = [self.visit(stmt) for stmt in preamble]
        return Module(preamble + module.body, ())

    def visit_Comment(self, node) -> AST:
        self.comments.append(node.text)
        return Expr(Constant(...))

    def visit_Builtin(self, node) -> AST:
        name = load(node.id)
        return self.visit(name)

    def visit_Symbol(self, node) -> AST:
        return self.require(node.value)

    def visit_Static(self, node) -> AST:
        if node.name is None:
            name = "_static_%s" % str(id(node.value)).replace('-', '_')
        else:
            name = node.name
        node = self.define(name, node.value)
        return self.visit(node)



from chameleon.exc import TemplateError
from chameleon.zpt.loader import TemplateLoader as PageTemplateLoader
from chameleon.zpt.template import PageTemplate
from chameleon.zpt.template import PageTemplateFile
from chameleon.zpt.template import PageTextTemplate
from chameleon.zpt.template import PageTextTemplateFile
